"""SAML TypedDict definitions for Flask-AppBuilder."""

from __future__ import annotations

from typing import Any, Dict, List, TypedDict


class SAMLServiceBinding(TypedDict):
    url: str
    binding: str


class SAMLIdPConfig(TypedDict, total=False):
    entityId: str
    singleSignOnService: SAMLServiceBinding
    singleLogoutService: SAMLServiceBinding
    x509cert: str


class SAMLProvider(TypedDict, total=False):
    name: str
    icon: str
    idp: SAMLIdPConfig
    attribute_mapping: Dict[str, str]


class SAMLSPConfig(TypedDict, total=False):
    entityId: str
    assertionConsumerService: SAMLServiceBinding
    singleLogoutService: SAMLServiceBinding
    NameIDFormat: str
    x509cert: str
    privateKey: str


class SAMLSecurityConfig(TypedDict, total=False):
    nameIdEncrypted: bool
    authnRequestsSigned: bool
    logoutRequestSigned: bool
    logoutResponseSigned: bool
    signMetadata: bool
    wantMessagesSigned: bool
    wantAssertionsSigned: bool
    wantAssertionsEncrypted: bool
    wantNameId: bool
    wantNameIdEncrypted: bool
    wantAttributeStatement: bool


class SAMLConfig(TypedDict, total=False):
    strict: bool
    debug: bool
    sp: SAMLSPConfig
    idp: SAMLIdPConfig
    security: SAMLSecurityConfig


class SAMLFlaskRequest(TypedDict):
    https: str
    http_host: str
    server_port: str
    script_name: str
    get_data: Dict[str, Any]
    post_data: Dict[str, Any]
    query_string: str


class SAMLUserInfo(TypedDict, total=False):
    username: str
    email: str
    first_name: str
    last_name: str
    role_keys: List[str]



import datetime
import logging
import re

from flask_sqlalchemy import (
    _QueryProperty,
    DefaultMeta,
    get_state,
    SessionBase,
    SignallingSession,
    SQLAlchemy,
)
from sqlalchemy import orm

try:
    from sqlalchemy.ext.declarative import as_declarative
except ImportError:
    from sqlalchemy.ext.declarative.api import as_declarative

try:
    from sqlalchemy.orm.util import identity_key  # noqa

    has_identity_key = True
except ImportError:
    has_identity_key = False

log = logging.getLogger(__name__)

_camelcase_re = re.compile(r"([A-Z]+)(?=[a-z0-9])")


class CustomSignallingSession(SignallingSession):
    """
    Custom Signaling Session to support SQLALchemy>=1.4 with flask-sqlalchemy 2.X
    https://github.com/pallets/flask-sqlalchemy/issues/953
    """

    def get_bind(self, mapper=None, *args, **kwargs):
        """Return the engine or connection for a given model or
        table, using the ``__bind_key__`` if it is set.

        Patch from https://github.com/pallets/flask-sqlalchemy/pull/1001
        """
        # mapper is None if someone tries to just get a connection
        if mapper is not None:
            try:
                # SA >= 1.3
                persist_selectable = mapper.persist_selectable
            except AttributeError:
                # SA < 1.3
                persist_selectable = mapper.mapped_table
            info = getattr(persist_selectable, "info", {})
            bind_key = info.get("bind_key")
            if bind_key is not None:
                state = get_state(self.app)
                return state.db.get_engine(self.app, bind=bind_key)
        return SessionBase.get_bind(self, mapper, *args, **kwargs)


class SQLA(SQLAlchemy):
    """
    This is a child class of flask_SQLAlchemy
    It's purpose is to override the declarative base of the original
    package. So that it is bound to F.A.B. Model class allowing the dev
    to be in the same namespace of the security tables (and others)
    and can use AuditMixin class alike.

    Use it and configure it just like flask_SQLAlchemy
    """

    def make_declarative_base(self, model, metadata=None):
        base = Model
        base.query = _QueryProperty(self)
        return base

    def get_tables_for_bind(self, bind=None):
        """Returns a list of all tables relevant for a bind."""
        result = []
        tables = Model.metadata.tables
        for key in tables:
            if tables[key].info.get("bind_key") == bind:
                result.append(tables[key])
        return result

    def create_session(self, options):
        """
        Custom Session factory to support SQLALchemy>=1.4 with flask-sqlalchemy 2.X

        https://github.com/pallets/flask-sqlalchemy/issues/953

        :param options: dict of keyword arguments passed to session class
        """

        return orm.sessionmaker(class_=CustomSignallingSession, db=self, **options)


class ModelDeclarativeMeta(DefaultMeta):
    """
    Base Model declarative meta for all Models definitions.
    Setups bind_keys to support multiple databases.
    Setup the table name based on the class camelcase name.
    """


@as_declarative(name="Model", metaclass=ModelDeclarativeMeta)
class Model(object):
    """
    Use this class has the base for your models,
    it will define your table names automatically
    MyModel will be called my_model on the database.

    ::

        from sqlalchemy import Integer, String
        from flask_appbuilder import Model

        class MyModel(Model):
            id = Column(Integer, primary_key=True)
            name = Column(String(50), unique = True, nullable=False)

    """

    __table_args__ = {"extend_existing": True}

    def to_json(self):
        result = dict()
        for key in self.__mapper__.c.keys():
            col = getattr(self, key)
            if isinstance(col, datetime.datetime) or isinstance(col, datetime.date):
                col = col.isoformat()
            result[key] = col
        return result



"""SAML utility helpers for Flask-AppBuilder."""

from __future__ import annotations

import logging
from typing import Any, Dict, List, Optional

from .types import SAMLUserInfo

log = logging.getLogger(__name__)


def map_saml_attributes(
    saml_attributes: Dict[str, List[str]],
    attribute_mapping: Dict[str, str],
    name_id: Optional[str] = None,
) -> SAMLUserInfo:
    """Map SAML assertion attributes to FAB user fields.

    Args:
        saml_attributes: Raw attributes from the SAML assertion.
        attribute_mapping: Mapping of SAML attribute names to FAB field names.
        name_id: The SAML NameID value (used as fallback for username/email).

    Returns:
        Dictionary with FAB user field names as keys.
    """
    userinfo: Dict[str, Any] = {}

    for saml_attr, fab_field in attribute_mapping.items():
        value = saml_attributes.get(saml_attr)
        if value is None:
            continue
        if fab_field == "role_keys":
            userinfo[fab_field] = list(value)
        else:
            userinfo[fab_field] = value[0] if value else ""

    # Fallback to NameID if no username/email mapped
    if name_id and "username" not in userinfo:
        userinfo["username"] = name_id
        if "@" in name_id and "email" not in userinfo:
            userinfo["email"] = name_id

    return userinfo


def fetch_idp_metadata(url: str) -> str:
    """Fetch IdP metadata XML from a remote URL."""
    import requests

    resp = requests.get(url, timeout=10)
    resp.raise_for_status()
    return resp.text



"""SP metadata generation for SAML authentication."""

from __future__ import annotations

import logging
from typing import Any, Dict

log = logging.getLogger(__name__)


def get_sp_metadata(saml_settings: Dict[str, Any]) -> str:
    """
    Generate SP metadata XML from SAML settings.

    :param saml_settings: The python3-saml settings dict
    :return: SP metadata XML string
    """
    from onelogin.saml2.settings import OneLogin_Saml2_Settings

    settings = OneLogin_Saml2_Settings(saml_settings, sp_validation_only=True)
    metadata = settings.get_sp_metadata()
    errors = settings.validate_metadata(metadata)
    if errors:
        log.error("SP Metadata validation errors: %s", errors)
        raise ValueError(f"SP Metadata validation errors: {errors}")
    return metadata
